/*
 *
 *    eth_lib3.c 
 *
 *    The routines in this file are an interface to the eht_hook schar driver
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h> 
#include <netinet/if_ether.h> 
#include <netinet/in.h> 
#include <netinet/ip.h> 
#include <net/if.h> 
#include <sys/ioctl.h> 
#include <netpacket/packet.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include "eth.h"

int  eth_open(char *dev_name);
     //opens network device and gets MAC address.

int  eth_register_mac(unsigned char *mac);
     //can register multiple MAC addresses one at a time.  The external
     //variable 'cur_mac' specifies which MAC will be used when 'eth_write'
     //is called.

void eth_close(void);
     //closes the device.

int  eth_reset(void);
     //reset driver's ring buffer pointers and stats.

int  eth_read();
     //gets the next available packet or timeout.
     //the data is stored in the extern array rbuf[].
     //the number of bytes returned is in nrbuf.

int  eth_write();
     //writes nwbuf bytes of data pointed to by wbuf to the registered
     //MAC address stored in the extern variable 'cur_mac'.
     //Note wbuf is only a pointer which points to the start of data in
     //the extern array ebuf[].

int  eth_read_timeout(int rd_tmo);
     //Sets the desired timeout for blocking reads.
     //The units are jiffies which are implementation dependent, but
     //are typically supposed to be one hundreth of a second.

int eth_get_pkts_left();
     //Reads the number of packets waiting to be read.


extern unsigned char ebuf[]; //write packet data including header. 
extern unsigned char *wbuf;  //pointer to start of data in ebuf[].
extern unsigned char rbuf[]; //read packt data including header.
extern int nwbuf;            //number of data bytes to write (excludes header)
extern int nrbuf;            //number of data bytes returned (inc. header)
extern struct ethhdr *ether_header[];//Dest. MAC, Source MAC, and Len
extern unsigned char hw_source_addr[]; //MAC of NIC in use
extern int nmacs;    //number of registered MAC addresses.
extern int cur_mac;  //index to current MAC device to write to. 


static int fd_schar; // file descriptor of schar driver.

#ifdef SIM_OUTPUT
static  FILE *smfp;
#endif
static char *fmt;
static char dsh[] = "%02X-"; 
static char nodsh[] = "%02X\n";

int eth_open(char *dev_name)
{
  int i;
  char *dev;
  if((dev=(char *)malloc((6+strlen(dev_name))*sizeof(char)))==0){
    printf("Device name allocation error!\n");
    return -1;
  }
  sprintf(dev,"/dev/%s",dev_name);
  fd_schar = open(dev, O_RDWR);
  if (fd_schar == -1) {
    perror("open");
    return -1;
  }
  printf("Opened network path on device %s\n",dev);
  free(dev);
  if(ioctl(fd_schar,SCHAR_GET_MAC,hw_source_addr)){
    printf("Error reading MAC address\n");
  }
  printf("NIC MAC addr.  (source): ");
  for(i=0;i<6;i++){
    fmt = (i<5)?dsh:nodsh;
    printf(fmt,hw_source_addr[i]&0xff);
  }
  nmacs=0;
#ifdef SIM_OUTPUT
  smfp=fopen("sim_out.hex","w");
#endif
}

int eth_register_mac(unsigned char *mac)
{
  int  i,mac_idx;
  mac_idx = nmacs++; 
  if((ether_header[mac_idx]=(struct ethhdr *)malloc(sizeof(struct ethhdr)))==0){
    printf("Could not register MAC address: Allocation error\n");
    return -1;
  }
  memcpy(ether_header[mac_idx]->h_dest, mac, ETH_ALEN);
  memcpy(ether_header[mac_idx]->h_source,hw_source_addr, ETH_ALEN);
  printf("Registered Dest. MAC %2d: ",mac_idx);
  for(i=0;i<6;i++){
    fmt = (i<5)?dsh:nodsh;
    printf(fmt,ether_header[mac_idx]->h_dest[i]&0xff);
  }
  return mac_idx;
}

int eth_reset(void)
{ 
  if(ioctl(fd_schar,SCHAR_RESET)){
    printf("Error in SCHAR_RESET \n");
  }
  return 0;
}

int eth_read()
{
int size;
int lp;
  do {
    size=read(fd_schar,rbuf,nrbuf);
    for(lp=0;size>=0 && size <7 && rbuf[0]==0x03 && lp<100;lp++){
      usleep(100);
      size=read(fd_schar,rbuf,nrbuf);
    }
  } while(size>6 && (rbuf[0]&1)==1); //ignore multicast packets
  return size;
}


int eth_write()
{
  int pkt_size,pkt_stp;
  int i,n_written;
  static  unsigned int A = 0x7a30ec8; 
  static  unsigned int B = 0x39b50F1;
  //  check for registered MAC address
  if(cur_mac<0 || cur_mac >= nmacs){
    printf("Your using an unregister MAC address\n");
    return -99;
  }
  //Creating the packet
  ether_header[cur_mac]->h_proto = htons(nwbuf);/* Length of data */
  memcpy(ebuf, ether_header[cur_mac], ETH_HLEN);
  pkt_size = ETH_HLEN + nwbuf;
  //  memcpy(ebuf+ETH_HLEN, wbuf, nwbuf); /*don't need this extra copy now*/
#ifdef SIM_OUTPUT
#ifdef SIM_TRUNK_MAX
  pkt_stp = SIM_TRUNK_MAX;
#else
  pkt_stp = pkt_size;
#endif
  for(i=0;i<pkt_size&&i<pkt_stp;i++)fprintf(smfp,"%02X",ebuf[i]&0xff);
#ifdef SIM_TRUNK_MAX
  if(i== SIM_TRUNK_MAX)  fprintf(smfp,"...");
#endif
  fprintf(smfp,"\n");
  fflush(smfp);
#endif
  n_written = write(fd_schar, (const void *)ebuf, pkt_size);
  //  for(i=0;i<100*nwbuf;i++)A=(A^B)>>1;//Slow down rate proportionally to size 
  return n_written;
}


void eth_close(void)
{
  int i;
#ifdef SIM_OUTPUT
  fclose(smfp);
#endif
  for(i=0;i<nmacs;i++){
    free(ether_header[i]);
  }
  close(fd_schar);
}

int eth_read_timeout(int rd_tmo)
{
  if(ioctl(fd_schar,SCHAR_READ_TIMEOUT,rd_tmo)){
    printf("Error in SCHAR_READ_TIMEOUT \n");
  }
  return 0;
}
int eth_get_pkts_left(void)
{
  return ioctl(fd_schar,SCHAR_PKTS_LEFT);
}
